''
''
''     QuadRotor - Full Stability
'' -- Jason Dorie --               
''               
''

{
  TODO:
        Use integrated gyro value for P, current gyro measurement for D
        Use accelerometer to bias integrated gyro result
                Store last 128 results in HUB RAM, keep a running sum, SHR 7 for current value

  DONE:
        Temperature drift compensation
        Convert ITG-3200 code to PASM
        Add ADXL345 readings to ITG code
        Integration of gyro values
}


CON
  _clkmode = xtal1 + pll8x
  _xinfreq = 10_000_000

  UpdateFreq = 200
  CorrectionDivisor = 256


OBJ
  Gyro  : "9DOF-pasm.spin"                              ' 1 cog
  'Flt   : "Float32Full-Limited_Routines-1_Cog"          ' 1 cog                        
  Dbg   : "FullDuplexSerial"                            ' 1 cog
  Xmt   : "FullDuplexSerial"                            ' 1 cog

  'Main loop                                            ' 1 cog

  DCM   : "PropDCM28.spin"                              ' 0 cogs
                          

VAR
  long counter
  long LastTime, CurTime

  long dGx, dGy, dGz                                    ' Working variables used by the DCM code
  long dAx, dAy, dAz

  long XAngle, YAngle, ZAngle

  long stack[100]



PUB Main | v, x , y

  'waitcnt(clkfreq*3+cnt)
  Xmt.Start( 18, 17, 0, 115200 )
  Dbg.Start( 31, 30, 0, 115200 )
  Gyro.Start( 20, 21 )
  'Flt.Start
  cognew(TransmitThread, @stack)

  RunDCM  'Can be put into its own cog if desired




PUB RunDCM | tx, tz, loopTime

  DCM.Init

  '{
  'Start by gathering an average accelerometer value to use as "down"
  'Assumes the board is flat / level on startup
  dAx := 0
  dAy := 0
  dAz := 0
  loopTime := cnt

  repeat counter from 1 to 256
    dAx += Gyro.GetAX
    dAy += Gyro.GetAY
    dAz += Gyro.GetAZ

    waitcnt( constant(80_000_000 / 200) + loopTime )
    loopTime += constant(80_000_000 / 200)

  dAx := dAx << (DCM#FixedBits - (8 + 8))               '8 bits of accel value, 8 bits of loop accumulation
  dAy := dAy << (DCM#FixedBits - (8 + 8))
  dAz := dAz << (DCM#FixedBits - (8 + 8))
  

  'BiasCorrect SHOULD NOT include 1G gravity
  DCM.SetBiasCorrect( dAx, dAy, dAz - DCM#Fixed_One )
  '}


  DCM.SetGyroScale(UpdateFreq)
  DCM.SetCorrectionDivisor(CorrectionDivisor)
  
  loopTime := cnt

  repeat
    dGx := Gyro.GetRX
    dGy := Gyro.GetRY
    dGz := Gyro.GetRZ
    dAx := Gyro.GetAX
    dAy := Gyro.GetAY
    dAz := Gyro.GetAZ


    DCM.SetValues( dGx, dGy, dGz, dAx, dAy, dAz )

    'tx := cnt
    DCM.Update
    'tx := cnt - tx
    'dbg.dec( tx )
    'dbg.tx( 1 )    

    'Get the YAW vector in 2D, convert the direction of the user stick into a 2D vector as well
    'dot product will tell me how close and in which direction I am

    'Get the M[1,0] and M[1,2] components of the body matrix (Y components of the X and Z vectors)
    'Those are the SIN of the angle between the ground and the craft


    'It's worth noting that the range returned from ASIN is +/-(PI/2), or +/- 90 degrees
    'therefore the ASIN results will only be useful if the craft is within 89 degrees of "upright"
    'unless additional work is done to determine full range angular values.



    'Uncomment this block when you just want to see the rotation values
    {
    tx := Flt.ASin( Flt.FFloatDiv28( DCM.GetMatrixValue(5) ) )     'Convert to int, then divide by 1<<28
    tz := Flt.ASin( Flt.FFloatDiv28( DCM.GetMatrixValue(3) ) )     'Convert to int, then divide by 1<<28 

    XAngle := Flt.FRound( Flt.FMul( tx , constant(65536.0 / (PI / 2.0)) ) ) 
    ZAngle := Flt.FRound( Flt.FMul( tz , constant(-65536.0 / (PI / 2.0)) ) )

    'For heading, I want an actual angular value, so this returns me an int between 0 & 65535, where 0 is forward
    'Shifting it up 2 to allow for more bits of precision when accumulating the desired yaw value
    YAngle := Flt.FRound( Flt.FMul( Flt.Atan2( Flt.FFloat(DCM.GetMatrixValue(6)), Flt.FFloat(DCM.GetMatrixValue(8))), constant(32768.0 / PI) ) ) & 65535 

    dbg.tx(1)
    dbg.hex( XAngle, 4 )
    dbg.tx(32)
    dbg.hex( YAngle, 4 )
    dbg.tx(32)
    dbg.hex( ZAngle, 4 )
    dbg.tx(11)
    '}

    'Debug print the accelerometer values
    {
    dbg.tx(1)
    dbg.hex( dAx, 4 )
    dbg.tx(32)
    dbg.hex( dAy, 4 )
    dbg.tx(32)
    dbg.hex( dAz, 4 )
    dbg.tx(11)
    '}

    'Uncomment for the DCM Visualizer with LocalMode = true, or with AccelCalibrate = true
    {
    dbg.tx( $77 )
    dbg.tx( $77 )
    dbg.tx( dGx >> 8 )
    dbg.tx( dGx )
    dbg.tx( dGy >> 8 )
    dbg.tx( dGy )
    dbg.tx( dGz >> 8 )
    dbg.tx( dGz )
    dbg.tx( dAx >> 8 )
    dbg.tx( dAx )
    dbg.tx( dAy >> 8 )
    dbg.tx( dAy )
    dbg.tx( dAz >> 8 )
    dbg.tx( dAz )
    '}     

    'Uncomment for the DCM Visualizer with LocalMode = false
    {
    dbg.tx( $78 )
    dbg.tx( $78 )

    dbg.tx( DCM.GetMatrixValue(0) >> 22 )
    dbg.tx( DCM.GetMatrixValue(0) >> 14 )
    dbg.tx( DCM.GetMatrixValue(1) >> 22 )
    dbg.tx( DCM.GetMatrixValue(1) >> 14 )
    dbg.tx( DCM.GetMatrixValue(2) >> 22 )
    dbg.tx( DCM.GetMatrixValue(2) >> 14 )

    dbg.tx( DCM.GetMatrixValue(3) >> 22 )
    dbg.tx( DCM.GetMatrixValue(3) >> 14 )
    dbg.tx( DCM.GetMatrixValue(4) >> 22 )
    dbg.tx( DCM.GetMatrixValue(4) >> 14 )
    dbg.tx( DCM.GetMatrixValue(5) >> 22 )
    dbg.tx( DCM.GetMatrixValue(5) >> 14 )

    'The last row is now computed in the vis tool
    }
    'TransmitMatrix
    loopTime += constant(80_000_000 / UpdateFreq)
    waitcnt( loopTime )

DAT
  ' Look-up table for rotation from ZXY to XYZ
  zxy2xyz long
  long 8, 6, 7
  long 2, 0, 1
  long 5, 3, 4

PUB TransmitThread
  repeat
    TransmitMatrix
    PrintMatrix
    waitcnt(clkfreq/20+cnt)

PUB TransmitMatrix | i, val
    repeat i from 0 to 8
      val := DCM.GetMatrixValue(zxy2xyz[i])
      val := (val + (1 << 20)) ~> 21
      if val > 127
        val := 127
      if val < -128
        val := -128
      xmt.tx(val)
    xmt.tx(13)

PUB PrintMatrix | i, j, val, sign, divisor, digit
    dbg.tx(0)
    repeat i from 0 to 8
      val := DCM.GetMatrixValue(i)
      if val => 0
        sign := 0
      else
        sign := 1
        val := -val
      val := (val + 1342177) / 2684355
      ifnot val
        sign := 0
      if sign
        dbg.tx("-")
      else
        dbg.tx(" ")
      divisor := 100
      repeat j from 1 to 3
        digit := val/divisor
        val //= divisor
        divisor /= 10
        dbg.tx(digit+"0")
        if j == 1
          dbg.tx(".")
      if i // 3 == 2
        dbg.tx(13)
      else
        dbg.tx(" ")
       